"""Pydantic Version of the AaC Definition Class."""
from uuid import UUID
from typing import Optional
from pydantic import BaseModel

from aac.in_out.files.aac_file import AaCFile
from aac.context.definition import Definition


class DefinitionModel(BaseModel):
    """REST API Model for the Definition class."""
    uid: Optional[UUID]
    name: str
    package: str
    is_user_editable: Optional[bool]
    content: Optional[str]
    source_uri: str
    structure: dict


def to_definition_model(definition: Definition) -> DefinitionModel:
    """Return a DefinitionModel representation from a Definition object."""
    return DefinitionModel(
        uid=definition.uid,
        name=definition.name,
        package=definition.package,
        is_user_editable=definition.source.is_user_editable,
        content=definition.content,
        source_uri=definition.source.uri,
        structure=definition.structure,
    )


def to_definition_class(definition_model: DefinitionModel) -> Definition:
    """Return a Definition object from a DefinitionModel object."""
    source = AaCFile(definition_model.source_uri, True, False)
    definition = Definition(name=definition_model.name, package=definition_model.package, content="", source=source, structure=definition_model.structure)
    definition.uid = definition_model.uid
    definition.content = definition.to_yaml()
    return definition



"""__init__.py module for the REST API plugin."""

# WARNING - DO NOT EDIT - YOUR CHANGES WILL NOT BE PROTECTED.
# This file is auto-generated by the aac gen-plugin and may be overwritten.

from os.path import join, dirname

from aac.context.language_context import LanguageContext
from aac.execute import hookimpl
from aac.execute.aac_execution_result import ExecutionResult, ExecutionStatus
from aac.execute.plugin_runner import PluginRunner

from rest_api.rest_api_impl import plugin_name, rest_api, gen_openapi_spec


rest_api_aac_file_name = "rest_api.aac"


def run_rest_api(host: str = "0.0.0.0", port: int = 8000) -> ExecutionResult:
    """
    Start a RESTful interface for interacting with and managing AaC.

    Args:
        host (str): Set the hostname of the service. Useful for operating behind proxies.
        port (int): The port to which the RESTful service will be bound.

    Returns:
         The results of the execution of the plugin rest-api command.
    """

    result = ExecutionResult(plugin_name, "rest-api", ExecutionStatus.SUCCESS, [])

    rest_api_result = rest_api(host, port)
    if not rest_api_result.is_success():
        return rest_api_result
    else:
        result.add_messages(rest_api_result.messages)

    return result


def run_gen_openapi_spec(output_directory: str) -> ExecutionResult:
    """
    Write the OpenAPI schema to a JSON file.

    Args:
         output_directory (str): The output directory in which to write the AaC OpenAPI JSON file.

    Returns:
         The results of the execution of the plugin gen-openapi-spec command.
    """

    result = ExecutionResult(
        plugin_name, "gen-openapi-spec", ExecutionStatus.SUCCESS, []
    )

    gen_openapi_spec_result = gen_openapi_spec(output_directory)
    if not gen_openapi_spec_result.is_success():
        return gen_openapi_spec_result
    else:
        result.add_messages(gen_openapi_spec_result.messages)

    return result


@hookimpl
def register_plugin() -> None:
    """
    Registers information about the plugin for use in the CLI.
    """

    active_context = LanguageContext()
    rest_api_aac_file = join(dirname(__file__), rest_api_aac_file_name)
    definitions = active_context.parse_and_load(rest_api_aac_file)

    rest_api_plugin_definition = [
        definition for definition in definitions if definition.name == plugin_name
    ][0]

    plugin_instance = rest_api_plugin_definition.instance
    for file_to_load in plugin_instance.definition_sources:
        active_context.parse_and_load(file_to_load)

    plugin_runner = PluginRunner(plugin_definition=rest_api_plugin_definition)
    plugin_runner.add_command_callback("rest-api", run_rest_api)
    plugin_runner.add_command_callback("gen-openapi-spec", run_gen_openapi_spec)

    active_context.register_plugin_runner(plugin_runner)



"""Pydantic Version of the AaC command Class."""
from pydantic import BaseModel

from aac.execute.plugin_runner import AacCommand, AacCommandArgument


class CommandRequestModel(BaseModel):
    """Command model for user requests to execute commands."""
    name: str
    arguments: list[str]


class CommandResponseModel(BaseModel):
    """Response model for returning the results/outcomes of command requests."""
    command_name: str
    result_message: str
    success: bool


class CommandArgumentModel(BaseModel):
    """REST API Model for the AaC Command Argument class."""
    name: str
    description: str
    data_type: str
    optional: bool


class CommandModel(BaseModel):
    """REST API Model for the AaC Command class. Used to present available commands."""
    name: str
    description: str
    arguments: list[CommandArgumentModel]


def to_command_argument_model(command_argument: AacCommandArgument) -> CommandArgumentModel:
    """Return a CommandModel representation from an AacCommand object."""
    return CommandArgumentModel(
        name=command_argument.name, description=command_argument.description, data_type=command_argument.data_type, optional=command_argument.name.startswith("--")
    )


def to_command_model(command: AacCommand) -> CommandModel:
    """Return a CommandModel representation from an AacCommand object."""
    command_argument_models = list(map(to_command_argument_model, command.arguments))
    return CommandModel(
        name=command.name, description=command.description, arguments=command_argument_models
    )



"""Pydantic Version of the AaC Definition Class."""
from typing import Optional
from pydantic import BaseModel, FilePath

from aac.in_out.files.aac_file import AaCFile


class FileModel(BaseModel):
    """REST API Model for the File class."""
    uri: str
    content: Optional[str]
    is_user_editable: bool
    is_loaded_in_context: bool


class FilePathModel(BaseModel):
    """REST API Model for just file uris."""
    uri: FilePath


class FilePathRenameModel(BaseModel):
    """REST API Model for renaming a file."""
    current_file_uri: FilePath
    new_file_uri: str


def to_file_model(file: AaCFile, file_content: Optional[str] = None) -> FileModel:
    """
    Return a FileModel representation from an AaCFile object.

    Args:
        file (AaCFile): The AaCFile object to convert
        file_content (str): The content of the file -- this information is not stored in AaCFile

    Returns:
        A FileModel for use in the REST API.
    """
    return FileModel(
        uri=file.uri, content=file_content, is_user_editable=file.is_user_editable, is_loaded_in_context=file.is_loaded_in_context
    )


def to_file_class(file: FileModel) -> AaCFile:
    """
    Return an AacFile object from a FileModel object.

    Args:
        file (FileModel): The FileModel object to convert

    Returns:
        An AaCFile object.
    """
    return AaCFile(
        uri=file.uri, is_user_editable=file.is_user_editable, is_loaded_in_context=file.is_loaded_in_context
    )



"""The AaC REST API plugin implementation module."""

# NOTE: It is safe to edit this file.
# This file is only initially generated by aac gen-plugin, and it won't be overwritten if the file already exists.

# There may be some unused imports depending on the definition of the plugin, be sure to remove unused imports.
import asyncio
import json
import logging
import os
import uvicorn
from fastapi.openapi.utils import get_openapi

from aac.execute.aac_execution_result import (
    ExecutionResult,
    ExecutionStatus,
    ExecutionMessage,
    MessageLevel,
)

from rest_api.aac_rest_app import app, refresh_available_files_in_workspace

plugin_name = "REST API"


def rest_api(host: str, port: int) -> ExecutionResult:
    """
    Business logic for allowing rest-api command to perform Start a RESTful interface for interacting with and managing AaC.

    Args:
         host (str): Set the hostname of the service. Useful for operating behind proxies.
         port (int): The port to which the RESTful service will be bound.

    Returns:
         The results of the execution of the rest-api command.
    """

    status = ExecutionStatus.GENERAL_FAILURE

    if isinstance(port, str):
        port = int(port)

    msg_str = _start_restful_service(host, port)

    status = ExecutionStatus.SUCCESS
    messages: list[ExecutionMessage] = []
    msg = ExecutionMessage(
        msg_str,
        MessageLevel.INFO,
        None,
        None,
    )
    messages.append(msg)

    return ExecutionResult(plugin_name, "rest-api", status, messages)


def _start_restful_service(host: str, port: int) -> str:
    """
    Starts the RESTful interface service.

    Args:
        host (str): The hostname of the service.
        port (int): The port to which the RESTful service will be bound.

    Returns:
        A success message.
    """
    asyncio.run(refresh_available_files_in_workspace())
    logging.info(f"Starting REST API in {os.getcwd()}.")
    uvicorn.run(app, host=host, port=port)
    return "Successfully started the RESTful API."


def gen_openapi_spec(output_directory: str) -> ExecutionResult:
    """
    Business logic for allowing gen-openapi-spec command to perform Write the OpenAPI schema to a JSON file.

    Args:
         output_directory (str): The output directory in which to write the AaC OpenAPI JSON file.

    Returns:
         The results of the execution of the gen-openapi-spec command.
    """

    status = ExecutionStatus.GENERAL_FAILURE

    msg_str, status = _write_openapi_spec_to_file(output_directory)

    messages: list[ExecutionMessage] = []
    msg = ExecutionMessage(
        msg_str,
        MessageLevel.INFO,
        None,
        None,
    )
    messages.append(msg)

    return ExecutionResult(plugin_name, "gen-openapi-spec", status, messages)


def _write_openapi_spec_to_file(output_directory: str) -> [str, ExecutionStatus]:
    """
    Writes the open api spec to a json file.

    Args:
        output_directory (str): The directory in which the json file will be created.

    Returns:
        A success message.
    """
    if not os.path.exists(output_directory):
        os.mkdir(output_directory)

    full_file_path = os.path.join(output_directory, "AaC_OpenAPI_Schema.json")

    try:
        with open(full_file_path, "w") as output_file:
            json.dump(
                get_openapi(
                    title=app.title,
                    version=app.version,
                    openapi_version=app.openapi_version,
                    description=app.description,
                    routes=app.routes,
                ),
                output_file,
                indent=4,
            )

        msg = f"Successfully wrote the OpenAPI spec to {full_file_path}."
        status = ExecutionStatus.SUCCESS
    except TypeError:
        msg = f"Unable to convert the OpenAPI spec to JSON: {TypeError}"
        status = ExecutionStatus.GENERAL_FAILURE

    return msg, status



from fastapi import FastAPI, HTTPException, BackgroundTasks, responses, exceptions, Request
from http import HTTPStatus
import logging
from typing import Optional
import os

from aac.context.definition_parser import DefinitionParser
from aac.context.language_context import LanguageContext
from aac.context.language_error import LanguageError
from aac.context.definition import Definition
from aac.in_out.files.aac_file import AaCFile
from aac.in_out.files.find import find_aac_files, is_aac_file
from aac.in_out.paths import sanitize_filesystem_path
from aac.in_out.parser import parse, ParserError
from aac.execute.aac_execution_result import ExecutionStatus
from aac.execute.plugin_runner import AacCommand, AacCommandArgument

from rest_api.models.command_model import (CommandModel, CommandRequestModel, CommandResponseModel, to_command_model)
from rest_api.models.definition_model import DefinitionModel, to_definition_class, to_definition_model
from rest_api.models.file_model import FileModel, FilePathModel, FilePathRenameModel, to_file_model

app = FastAPI()

AVAILABLE_AAC_FILES: list[AaCFile] = []
WORKSPACE_DIR: str = os.getcwd()


# File CRUD Operations
def _get_files_in_context() -> list[AaCFile]:
    """
    Returns a list of all files contributing definitions to the active context.

    Returns:
        A list of all files contributing definitions to the active context.
    """

    active_context = LanguageContext()

    return list({definition.source for definition in active_context.get_definitions()})


def _get_file_in_context_by_uri(uri: str) -> Optional[AaCFile]:
    """
    Return the AaCFile object by uri from the context or None if the file isn't in the context.

    Args:
        uri (str): The string uri to search for.

    Returns:
        An optional AaCFile if it's present in the context, otherwise None.
    """

    active_context = LanguageContext()

    for definition in active_context.get_definitions():
        if definition.source.uri == uri:
            return definition.source


def _get_definitions_by_file_uri(file_uri: str) -> list[Definition]:
    """
    Return a subset of definitions that are sourced from the target file URI.

    Args:
        file_uri (str): The source file URI to filter on.

    Returns:
        A list of definitions belonging to the target file.
    """

    active_context = LanguageContext()
    definitions = active_context.get_definitions()
    return [definition for definition in definitions if str(file_uri) == str(definition.source.uri)]


@app.get("/files/context", status_code=HTTPStatus.OK, response_model=list[FileModel])
def get_files_from_context():
    """
    Return a list of file model definitions of files contributing definitions.

    Returns:
        A list of FileModel objects representing files in the active context.
    """
    return [to_file_model(file) for file in _get_files_in_context()]


@app.get("/files/available", status_code=HTTPStatus.OK, response_model=list[FileModel])
def get_available_files(background_tasks: BackgroundTasks):
    """
    Return a list of all files available in the workspace for import into the active context. The list of files returned does not include files already in the context.

    Args:
        background_tasks (BackgroundTasks): A BackgroundTasks object containing all currently active background tasks.

    Returns:
        A list of files available for import into the active context.
    """
    # Update the files via an async function so that any changes to the files shows up, eventually.
    background_tasks.add_task(refresh_available_files_in_workspace)

    #  Having to use a cached response for now as the file-walking makes the response take too long.
    return [to_file_model(file) for file in list(_get_available_files_in_workspace())]


@app.get("/file", status_code=HTTPStatus.OK, response_model=FileModel)
def get_file_by_uri(uri: str):
    """
    Return the target file from the workspace

    Args:
        uri (str): The string uri to search for.

    Returns:
        Target file from the workspace, or HTTPStatus.NOT_FOUND if the file isn't in the context
    """
    file_in_context = _get_file_in_context_by_uri(uri)

    if file_in_context:
        file_model = to_file_model(file_in_context)
        with open(file_in_context.uri) as file:
            file_model.content = file.read()

        return file_model
    else:
        _report_error_response(HTTPStatus.NOT_FOUND, f"File {uri} not found in the context.")


@app.post("/files/import", status_code=HTTPStatus.NO_CONTENT)
def import_files_to_context(file_models: list[FilePathModel]) -> None:
    """
    Import the list of files into the context.

    Args:
        file_models (list[FilePathModel]): List of file models for import.
    """

    active_context = LanguageContext()

    files_to_import = set([str(model.uri) for model in file_models])
    valid_aac_files = set(filter(is_aac_file, files_to_import))
    invalid_files = files_to_import.difference(valid_aac_files)
    if len(invalid_files) > 0:
        _report_error_response(
            HTTPStatus.BAD_REQUEST,
            f"Invalid files were asked to imported. Invalid files: {invalid_files}.",
        )
    else:
        try:
            new_file_definitions = [parse(file) for file in valid_aac_files]
        except ParserError as error:
            raise ParserError(error.source, error.errors) from None

        parser = DefinitionParser()
        for file in new_file_definitions:
            parser.load_definitions(active_context, file)


@app.put("/file", status_code=HTTPStatus.NO_CONTENT)
def rename_file_uri(rename_request: FilePathRenameModel) -> None:
    """
    Update a file's uri. (Rename file).

    Args:
        rename_request (FilePathRenameModel): A RestAPI model for renaming a file.
    """
    current_file_path = sanitize_filesystem_path(str(rename_request.current_file_uri))
    new_file_path = sanitize_filesystem_path(rename_request.new_file_uri)

    file_in_context = _get_file_in_context_by_uri(current_file_path)

    if not _is_file_path_in_working_directory(new_file_path):
        _report_error_response(
            HTTPStatus.BAD_REQUEST,
            f"Files can only be renamed to a uri inside of the working directory: {WORKSPACE_DIR}.",
        )

    if file_in_context:
        os.rename(current_file_path, new_file_path)
        definitions_to_update = _get_definitions_by_file_uri(current_file_path)
        for definition in definitions_to_update:
            definition.source.uri = new_file_path

    else:
        _report_error_response(HTTPStatus.NOT_FOUND, f"File {current_file_path} not found in the context.")


@app.delete("/file", status_code=HTTPStatus.NO_CONTENT)
def remove_file_by_uri(uri: str) -> None:
    """
    Remove the requested file and it's associated definitions from the active context.

    Args:
        uri (str): uri (str): The string uri of the files to be removed.
    """
    active_context = LanguageContext()

    file_in_context = _get_file_in_context_by_uri(uri)
    if not file_in_context:
        _report_error_response(HTTPStatus.NOT_FOUND, f"File {uri} not found in the context.")

    definitions_to_remove = []
    discovered_definitions = _get_definitions_by_file_uri(uri)
    definitions_to_remove.extend(discovered_definitions)

    if len(discovered_definitions) == 0:
        _report_error_response(
            HTTPStatus.NOT_FOUND,
            f"No definition(s) from {uri} were found in the context; Will not remove any definitions or files from the context.",
        )

    active_context.remove_definitions(definitions_to_remove)


# Definition CRUD Operations
@app.get("/definitions", status_code=HTTPStatus.OK, response_model=list[DefinitionModel])
def get_definitions() -> list[DefinitionModel]:
    """
    Return a list of the definitions in the active context.

    Returns:
        A list of definitions represented as DefinitionModel objects.
    """
    active_context = LanguageContext()

    definition_models = [to_definition_model(definition) for definition in active_context.get_definitions()]
    return definition_models


@app.get("/definition", status_code=HTTPStatus.OK, response_model=list[DefinitionModel])
def get_definition_by_name(name: str) -> list[DefinitionModel]:
    """
    Returns a definition from active context by name, or HTTPStatus.NOT_FOUND not found if the definition doesn't exist.

    Args:
        name (str): Name of the definition to be returned

    Returns:
        Returns the definitions with the given name as a list containing DefinitionModel objects.
    """
    active_context = LanguageContext()

    definitions = active_context.get_definitions_by_name(name)

    if not definitions:
        _report_error_response(HTTPStatus.NOT_FOUND, f"Definition {name} not found in the context.")
    else:
        definition_models = [to_definition_model(definition) for definition in definitions]

        return definition_models


@app.post("/definition", status_code=HTTPStatus.NO_CONTENT)
def add_definition(definition_model: DefinitionModel) -> None:
    """
    Add the definition to the active context. If the definition's source file doesn't exist, a new one will be created.

    Args:
        definition_model (DefinitionModel): The definition model in request body.

    Returns:
        204 HTTPStatus.NO_CONTENT if successful.
    """
    active_context = LanguageContext()

    definition_source_uri = sanitize_filesystem_path(definition_model.source_uri)

    if not _is_file_path_in_working_directory(definition_source_uri):
        _report_error_response(
            HTTPStatus.BAD_REQUEST,
            f"Definition can't be added to a file {definition_source_uri} which is outside of the working directory: {WORKSPACE_DIR}.",
        )
    definition_to_add = to_definition_class(definition_model)
    existing_definitions = _get_definitions_by_file_uri(definition_source_uri)

    is_user_editable = True
    if len(existing_definitions) > 0:
        is_user_editable = existing_definitions[0].source.is_user_editable

    if not is_user_editable:
        _report_error_response(
            HTTPStatus.BAD_REQUEST,
            f"File {definition_source_uri} can't be edited by users.",
        )

    parser = DefinitionParser()
    parser.load_definitions(active_context, [definition_to_add])


@app.post("/definitions", status_code=HTTPStatus.NO_CONTENT)
def add_definitions(definition_models: list[DefinitionModel]) -> None:
    """
    Add the definitions to the active context. If the definition's source file doesn't exist, a new one will be created.

    Args:
        definition_models (list[DefinitionModel]): The list of definition models in request body.

    Returns:
        204 HTTPStatus.NO_CONTENT if successful.
    """
    active_context = LanguageContext()

    definitions_to_add = []
    for definition_model in definition_models:
        definition_source_uri = sanitize_filesystem_path(definition_model.source_uri)

        if not _is_file_path_in_working_directory(definition_source_uri):
            _report_error_response(
                HTTPStatus.BAD_REQUEST,
                f"Definition can't be added to a file {definition_source_uri} which is outside of the working directory: {WORKSPACE_DIR}.",
            )
        definitions_to_add.append(to_definition_class(definition_model))
        existing_definitions = _get_definitions_by_file_uri(definition_source_uri)

        is_user_editable = True
        if len(existing_definitions) > 0:
            is_user_editable = existing_definitions[0].source.is_user_editable

        if not is_user_editable:
            _report_error_response(
                HTTPStatus.BAD_REQUEST,
                f"File {definition_source_uri} can't be edited by users.",
            )

    parser = DefinitionParser()
    parser.load_definitions(active_context, definitions_to_add)


@app.put("/definition", status_code=HTTPStatus.NO_CONTENT)
def update_definition(definition_model: DefinitionModel) -> None:
    """
    Update the request body definitions in the active context.

    Args:
        definition_model (DefinitionModel): The definition to be updated.
    """
    active_context = LanguageContext()

    definitions_to_update = active_context.get_definitions_by_name(definition_model.name)

    if definitions_to_update:
        updated_definition = to_definition_class(definition_model)
        for definition in definitions_to_update:
            if definition.name == updated_definition.name:
                updated_definition.uid = definition.uid
        active_context.remove_definitions(definitions_to_update)
        parser = DefinitionParser()
        parser.load_definitions(active_context, [updated_definition])

    else:
        _report_error_response(
            HTTPStatus.NOT_FOUND,
            f"Definition(s) {definition_model.name} not found in the context; failed to update definitions.",
        )


@app.delete("/definition", status_code=HTTPStatus.NO_CONTENT)
def remove_definition_by_name(name: str) -> None:
    """
    Remove the definition via name from the active context.

    Args:
        name (str): Name of the definition to be removed.
    """
    active_context = LanguageContext()

    definitions_to_remove = active_context.get_definitions_by_name(name)

    if definitions_to_remove:
        active_context.remove_definitions(definitions_to_remove)
    else:
        _report_error_response(
            HTTPStatus.NOT_FOUND,
            f"Definition {name} not found in the context; failed to delete definitions.",
        )


# Language Context Support

@app.get("/context/schema", status_code=HTTPStatus.OK, response_model=DefinitionModel)
def get_root_key_schema(key: str) -> DefinitionModel:
    """
    Returns the YAML schema for the given root key, or HTTPStatus.NOT_FOUND not found if the key doesn't exist.

    Args:
        key (str): The key of the root schema to be returned.

    Returns:
        Schema_Model containing the root schema with the given key.
        200 HTTPStatus.OK if successful.
        404 HTTPStatus.NOT_FOUND if the key doesn't exist.
    """
    active_context = LanguageContext()

    root_definitions = [definition for definition in active_context.get_definitions() if definition.get_root_key()]
    matching_definitions = [definition for definition in root_definitions if definition.name == key.capitalize()]

    if not matching_definitions:
        _report_error_response(HTTPStatus.NOT_FOUND, f"No root key found called {key}.")
    else:
        schema_definition = matching_definitions[0]

        if not schema_definition:
            _report_error_response(HTTPStatus.NOT_FOUND, f"Unable to get the schema definition {schema_definition.name}.")
        else:
            schema_model = to_definition_model(schema_definition)
            return schema_model


@app.get("/context/root_keys", status_code=HTTPStatus.OK, response_model=list[str])
def get_language_context_root_keys() -> list[str]:
    """
    Returns a list of root keys from the active context.

    Returns:
        A list containing all root keys in the active context.
        200 HTTPStatus.OK
    """
    active_context = LanguageContext()
    return [str(definition.get_root_key()) for definition in active_context.definitions if definition.get_root_key()]


# AaC Plugin Commands


@app.get("/commands", status_code=HTTPStatus.OK, response_model=list[CommandModel])
def get_aac_commands() -> list[CommandModel]:
    """
    Return a list of all available plugin commands.

    Returns:
        A list of CommandModel objects
    """
    aac_and_plugin_commands = _get_rest_api_compatible_commands()
    return [to_command_model(aac_and_plugin_commands[command]) for command in aac_and_plugin_commands]


@app.post("/command", status_code=HTTPStatus.OK, response_model=CommandResponseModel)
def execute_aac_command(command_request: CommandRequestModel):
    """
    Execute the command and return the result.

    Args:
        command_request (CommandRequestModel): The AaC command to be executed.

    Returns:
        A CommandResponseModel object containing the result of the command execution.
    """
    aac_commands_by_name = _get_rest_api_compatible_commands()
    aac_command = aac_commands_by_name.get(command_request.name)

    if aac_command is not None:
        aac_command_argument_names = [arg.name for arg in aac_command.arguments]
        arguments = [arg for arg in command_request.arguments if arg not in aac_command_argument_names]

        try:
            result = aac_command.callback(*(arguments or []))
            success = result.status_code == ExecutionStatus.SUCCESS
            result_message = f"{result.plugin_name}: {result.status_code.name.lower()}\n\n{result.get_messages_as_string()}"
        except Exception as error:
            success = False
            result_message = f"{result.plugin_name}: failure\n\n{error}"
        finally:
            return CommandResponseModel(command_name=aac_command.name, result_message=result_message, success=success)
    else:
        _report_error_response(
            HTTPStatus.NOT_FOUND,
            f"Command name {command_request.name} not found in the list of available commands: {list(aac_commands_by_name.keys())}.",
        )


def _get_available_files_in_workspace() -> set[AaCFile]:
    """
    Get the available AaC files in the workspace sans files already in the context.

    Returns:
        A set containing available AaC files in workspace.
    """
    aac_files_in_context = set(_get_files_in_context())
    aac_files_in_workspace = set(find_aac_files(WORKSPACE_DIR))

    return aac_files_in_workspace.difference(aac_files_in_context)


async def refresh_available_files_in_workspace() -> None:
    """
    Used to refresh the available files. Used in async since it takes too long for being used in request-response flow.
    """
    active_context = LanguageContext()

    AVAILABLE_AAC_FILES = list(_get_available_files_in_workspace())

    # Update the active context with any missing files
    files_in_context = {file.uri for file in _get_files_in_context()}
    available_files = {file.uri for file in AVAILABLE_AAC_FILES}
    missing_files = available_files.difference(files_in_context)
    try:
        definition_lists_from_missing_files = [parse(file_uri) for file_uri in missing_files]
    except ParserError as error:
        raise ParserError(error.source, error.errors) from None
    else:
        definitions_to_add = {definition.name: definition for definition_list in definition_lists_from_missing_files for definition in definition_list}
        parser = DefinitionParser()
        parser.load_definitions(active_context, list(definitions_to_add.values()))


def _report_error_response(code: HTTPStatus, error: str):
    """
    Accepts an error string and raises an HTTPException error.

    Args:
        code (HTTPStatus): The HTTP Status code
        error (str): An error message\

    Raises:
        HTTPException.
    """
    logging.error(error)
    raise HTTPException(
        status_code=code,
        detail=error,
    )


def _is_file_path_in_working_directory(file_path: str) -> bool:
    """
    Checks if the file path exists in the working directory.

    Args:
        file_path (str): Path to the file.

    Returns:
        A bool value of True if the file path exists in the working directory.
    """
    return str(file_path).startswith(WORKSPACE_DIR)


def _get_rest_api_compatible_commands() -> dict[str, AacCommand]:
    """
    Filter out plugin commands that aren't compatible with the rest-api command. These commands are long-running commands that don't allow for a timely rest response.

    Returns:
        A dictionary containing compatible commands, with the commands name as the key.
    """
    active_context = LanguageContext()

    long_running_commands = ["rest-api", "start-lsp-io", "start-lsp-tcp"]

    result: list[AacCommand] = []
    for runner in active_context.get_plugin_runners():
        definition = runner.plugin_definition
        for plugin_command in definition.instance.commands:
            if plugin_command not in long_running_commands:
                arguments: list[AacCommandArgument] = []
                for input in plugin_command.input:
                    arguments.append(
                        AacCommandArgument(
                            input.name,
                            input.description,
                            active_context.get_python_type_from_primitive(input.type),
                            input.default,
                        )
                    )
                result.append(
                    AacCommand(
                        plugin_command.name,
                        plugin_command.help_text,
                        runner.command_to_callback[plugin_command.name],
                        arguments,
                    )
                )

    return {command.name: command for command in result}


# Error Handlers


@app.exception_handler(LanguageError)
async def language_error_exception_handler(request, exc):
    """
    If a `LanguageError` exception is encountered, then return a 400 BAD Request with the exception's message.

    Args:
        request (Request): The encountered request.
        exc (LanguageError): The encountered LanguageError exception.

    Returns:
        The exception message response.
    """
    return responses.PlainTextResponse(str(exc), status_code=400)


@app.exception_handler(exceptions.RequestValidationError)
async def validation_exception_handler(request: Request, exc: exceptions.RequestValidationError):
    """
    If a `exceptions.RequestValidationError` exception is encountered, then return a 422 UNPROCESSABLE ENTITY with the exception's message.

    Args:
        request (Request): The encountered request.
        exc (RequestValidationError): The encountered RequestValidationError Exception.
    """
    _report_error_response(HTTPStatus.UNPROCESSABLE_ENTITY, str(exc))



import json
import os

from click.testing import CliRunner
from typing import Tuple
from tempfile import TemporaryDirectory
from unittest import TestCase

from fastapi.encoders import jsonable_encoder
from fastapi.testclient import TestClient
from http import HTTPStatus

from aac.context.constants import DEFINITION_FIELD_NAME
from aac.in_out.parser._parse_source import parse
from aac.execute.command_line import cli, initialize_cli
from rest_api.models.command_model import CommandRequestModel
from rest_api.models.definition_model import to_definition_model
from rest_api.models.file_model import FilePathModel, FilePathRenameModel
from rest_api.aac_rest_app import app, refresh_available_files_in_workspace


class TestRestApiCommands(TestCase):
    test_client = TestClient(app)

    def test_get_available_commands(self):
        response = self.test_client.get("/commands")
        self.assertEqual(HTTPStatus.OK, response.status_code)
        print(response.text)
        self.assertIn("check", response.text)
        self.assertIn("gen-plugin", response.text)
        self.assertIn("gen-project", response.text)
        self.assertIn("version", response.text)
        self.assertIn("clean", response.text)

    def test_execute_check_command(self):
        command_name = "check"
        test_model = parse(TEST_MODEL)[0]

        request_arguments = CommandRequestModel(name=command_name, arguments=[TEST_MODEL, "False", "False"])
        response = self.test_client.post("/command", data=json.dumps(jsonable_encoder(request_arguments)))

        self.assertEqual(HTTPStatus.OK, response.status_code)
        self.assertTrue(response.json().get("success"))
        self.assertIn("success", response.text)
        self.assertIn(command_name, response.text)
        self.assertIn(test_model.name, response.text)

    def test_execute_check_command_fails(self):
        command_name = "check"
        with self.assertRaises(Exception) as context:
            request_arguments = CommandRequestModel(name=command_name, arguments=[BAD_TEST_MODEL, "False", "False"])
            self.test_client.post("/command", data=json.dumps(jsonable_encoder(request_arguments)))


class TestAacRestApiFiles(TestCase):
    test_client = TestClient(app)

    def test_post_and_get_files(self):

        filepath = "tests/calc/model/calculator.yaml"
        self.assertTrue(os.path.isfile(filepath))

        file_model = [FilePathModel(uri=os.path.abspath(filepath))]
        self.test_client.post("/files/import", data=json.dumps(jsonable_encoder(file_model)))
        response = self.test_client.get("/files/context")
        self.assertEqual(HTTPStatus.OK, response.status_code)
        self.assertIn("calculator.yaml", response.text)

        available_files = self.test_client.get("/files/available")
        self.assertNotIn("calculator.yaml", available_files.text)

    def test_get_file_in_context_by_uri(self):
        filepath = "tests/calc/model/calculator.yaml"
        self.assertTrue(os.path.isfile(filepath))
        file_model = [FilePathModel(uri=os.path.abspath(filepath))]
        self.test_client.post("/files/import", data=json.dumps(jsonable_encoder(file_model)))

        response = self.test_client.get(f"/file?uri={os.path.abspath(filepath)}")
        self.assertIn(filepath, response.text)

    def test_rename_file_in_context(self):
        with TemporaryDirectory(dir=os.getcwd()) as temp_dir:
            old_file_name = "OldTestFile.yaml"
            new_file_name = "TestFile.aac"
            new_file_uri = os.path.join(temp_dir, new_file_name)

            temp_file_path = os.path.abspath(os.path.join(temp_dir, old_file_name))
            temp_file = open(temp_file_path, "w")
            temp_file.writelines(TEST_MODEL)
            temp_file.close()

            self.test_client.post("/files/import", data=json.dumps(jsonable_encoder([FilePathModel(uri=temp_file_path)])))
            rename_request_data = FilePathRenameModel(current_file_uri=temp_file_path, new_file_uri=new_file_uri)
            rename_response = self.test_client.put("/file", data=json.dumps(jsonable_encoder(rename_request_data)))
            self.assertEqual(HTTPStatus.NO_CONTENT, rename_response.status_code)

            response = self.test_client.get("/files/context")

            self.assertIn(new_file_uri, response.text)
            os.remove(new_file_uri)

    def test_remove_file_from_context(self):
        filepath = "tests/calc/model/calculator.yaml"
        self.assertTrue(os.path.isfile(filepath))
        file_model = [FilePathModel(uri=os.path.abspath(filepath))]
        self.test_client.post("/files/import", data=json.dumps(jsonable_encoder(file_model)))

        response = self.test_client.get("/files/context")
        self.assertEqual(HTTPStatus.OK, response.status_code)
        self.assertIn("calculator.yaml", response.text)

        result = self.test_client.delete(f"/file?uri={os.path.abspath(filepath)}")
        self.assertEqual(HTTPStatus.NO_CONTENT, result.status_code)

        response = self.test_client.get("/files/context")
        self.assertEqual(HTTPStatus.OK, response.status_code)
        self.assertNotIn("calculator.yaml", response.text)


class TestAacRestApiDefinitions(TestCase):
    test_client = TestClient(app)

    def test_get_definitions(self):

        filepath = "tests/calc/model/calculator.yaml"
        self.maxDiff = None

        model_path = os.path.abspath(filepath)
        self.assertTrue(os.path.isfile(model_path))
        definitions = parse(model_path)
        definition_model = to_definition_model(definitions[0])

        post_response = self.test_client.post("/definition", data=json.dumps(jsonable_encoder(definition_model)))
        self.assertEqual(HTTPStatus.NO_CONTENT, post_response.status_code)

        response = self.test_client.get("/definitions")
        self.assertEqual(HTTPStatus.OK, response.status_code)
        self.assertIn(filepath, response.text)

    def test_get_definition_by_name(self):
        refresh_available_files_in_workspace()
        filepath = "tests/calc/model/calculator.yaml"

        defs_to_lookup = ["Calculator", "Multiply", "MathLogger", "Add"]

        model_path = os.path.abspath(filepath)
        self.assertTrue(os.path.isfile(model_path))
        definitions = parse(model_path)
        definition_models = []
        for definition in definitions:
            definition_models.append(to_definition_model(definition))

        post_response = self.test_client.post("/definitions", data=json.dumps(jsonable_encoder(definition_models)))
        self.assertEqual(HTTPStatus.NO_CONTENT, post_response.status_code)

        for definition_name in defs_to_lookup:
            response = self.test_client.get(f"/definition?{DEFINITION_FIELD_NAME}={definition_name}")
            self.assertEqual(HTTPStatus.OK, response.status_code)
            self.assertIn(definition_name, response.text)

    def test_get_definition_by_name_not_found(self):
        fake_definition_name = "FakeModel"

        response = self.test_client.get(f"/definition/{fake_definition_name}")
        self.assertEqual(HTTPStatus.NOT_FOUND, response.status_code)

    def test_update_definitions(self):
        refresh_available_files_in_workspace()

        parsed_definition = parse(TEST_MODEL)[0]
        updated_parsed_definition = parse(UPDATED_TEST_MODEL)[0]

        post_response = self.test_client.post("/definition", data=json.dumps(jsonable_encoder(to_definition_model(parsed_definition))))
        self.assertEqual(HTTPStatus.NO_CONTENT, post_response.status_code)

        get_response = self.test_client.get(f"/definition?{DEFINITION_FIELD_NAME}={parsed_definition.name}")
        self.assertEqual(HTTPStatus.OK, get_response.status_code)
        self.assertIn("A TestModel", get_response.text)
        self.assertNotIn("An updated TestModel", get_response.text)

        update_response = self.test_client.put("/definition", data=json.dumps(jsonable_encoder(to_definition_model(updated_parsed_definition))))
        self.assertEqual(HTTPStatus.NO_CONTENT, update_response.status_code)

        get_response = self.test_client.get(f"/definition?{DEFINITION_FIELD_NAME}={parsed_definition.name}")
        self.assertEqual(HTTPStatus.OK, get_response.status_code)
        self.assertIn("An updated TestModel", get_response.text)
        self.assertNotIn("A TestModel", get_response.text)

    def test_remove_definition(self):
        refresh_available_files_in_workspace()
        filepath = "tests/calc/model/calculator.yaml"
        definition_to_be_deleted = "MathLogger"

        model_path = os.path.abspath(filepath)
        self.assertTrue(os.path.isfile(model_path))
        definitions = parse(model_path)
        definition_models = []
        for definition in definitions:
            definition_models.append(to_definition_model(definition))

        post_response = self.test_client.post("/definitions", data=json.dumps(jsonable_encoder(definition_models)))
        self.assertEqual(HTTPStatus.NO_CONTENT, post_response.status_code)

        get_response = self.test_client.get("/definitions")
        self.assertEqual(HTTPStatus.OK, get_response.status_code)
        self.assertIn("A log management service for calculator.", get_response.text)

        self.test_client.delete(f"/definition?{DEFINITION_FIELD_NAME}={definition_to_be_deleted}")

        get_response = self.test_client.get("/definitions")
        self.assertEqual(HTTPStatus.OK, get_response.status_code)
        self.assertNotIn("A log management service for calculator.", get_response.text)

    def test_get_schema_definition(self):
        get_response = self.test_client.get("/context/schema?key=model")
        self.assertEqual(HTTPStatus.OK, get_response.status_code)
        self.assertIn("Model", get_response.text)
        self.assertIn("A definition that represents a system and/or component model.", get_response.text)

    def test_get_root_keys(self):
        refresh_available_files_in_workspace()
        filepath = "tests/calc/model/calculator.yaml"

        model_path = os.path.abspath(filepath)
        self.assertTrue(os.path.isfile(model_path))
        definitions = parse(model_path)
        definition_models = []
        for definition in definitions:
            definition_models.append(to_definition_model(definition))

        post_response = self.test_client.post("/definitions", data=json.dumps(jsonable_encoder(definition_models)))
        self.assertEqual(HTTPStatus.NO_CONTENT, post_response.status_code)

        get_response = self.test_client.get("/context/root_keys")
        self.assertEqual(HTTPStatus.OK, get_response.status_code)
        self.assertIn("model", get_response.text)
        self.assertIn("plugin", get_response.text)
        self.assertIn("req", get_response.text)
        self.assertIn("req_spec", get_response.text)


class TestGenOpenApiSpec(TestCase):

    def test_gen_openapi_spec(self):
        # Like in core going to rely on the CLI testing for this, have not determined what we would like to test here
        pass

    def run_gen_openapi_spec_cli_command_with_args(
        self, args: list[str]
    ) -> Tuple[int, str]:
        """Utility function to invoke the CLI command with the given arguments."""
        initialize_cli()
        runner = CliRunner()
        result = runner.invoke(cli, ["gen-openapi-spec"] + args)
        exit_code = result.exit_code
        std_out = str(result.stdout)
        output_message = std_out.strip().replace("\x1b[0m", "")
        return exit_code, output_message

    def test_cli_gen_openapi_spec(self):
        """Test the gen-openapi-spec CLI command success for the RestAPI Plugin."""
        with TemporaryDirectory() as temp_dir:
            args = [temp_dir]
            exit_code, output_message = self.run_gen_openapi_spec_cli_command_with_args(args)

            temp_dir_files = os.listdir(temp_dir)
            self.assertNotEqual(0, len(temp_dir_files))

    def test_cli_gen_openapi_spec_output(self):
        with TemporaryDirectory() as temp_dir:
            args = [temp_dir]
            exit_code, output_message = self.run_gen_openapi_spec_cli_command_with_args(args)

            temp_dir_files = os.listdir(temp_dir)
            for temp_file in temp_dir_files:
                self.assertTrue(temp_file.find("_OpenAPI_Schema.json"))
                temp_file_content = open(os.path.join(temp_dir, temp_file), "r")
                temp_content = temp_file_content.read()
                self.assertIn("/files/context", temp_content)
                self.assertIn("/files/available", temp_content)
                self.assertIn("delete", temp_content)
                self.assertIn("Get Definition By Name", temp_content)


TEST_MODEL = """
model:
    name: TestModel
    description: A TestModel
"""

UPDATED_TEST_MODEL = """
model:
    name: TestModel
    description: An updated TestModel
"""

BAD_TEST_MODEL = """
model:
    description: A TestModel
"""
